import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.custom_hash.TObjectIntCustomHashMap;
import gnu.trove.map.hash.TIntDoubleHashMap;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.HashMap;
import java.util.Vector;

public class Graph
{
    private boolean directed;
    private String name;
    private TIntObjectHashMap<TIntHashSet>[] outAdjList;
    private TIntObjectHashMap<TIntHashSet>[] inAdjList;
    private int[] nodeListColors;
    private int numNodeCols;
    private int numEdgeCols;
    private Vector<TIntHashSet> listEdgeCols;
    private TObjectIntCustomHashMap<TIntHashSet> mapEdgeColsIds;
    private int maxMultiplicity;

    public Graph(boolean directed, int numNodes)
    {
        this.directed=directed;
        this.outAdjList=new TIntObjectHashMap[numNodes];
        this.inAdjList=new TIntObjectHashMap[numNodes];
        this.nodeListColors=new int[numNodes];
        this.listEdgeCols=new Vector<>();
        this.mapEdgeColsIds=new TObjectIntCustomHashMap<>(new IntHashSetStrategy());
        for(int i=0;i<numNodes;i++)
        {
            outAdjList[i] = new TIntObjectHashMap<>();
            inAdjList[i] = new TIntObjectHashMap<>();
        }
    }

    public Graph(Graph net)
    {
        this.directed=net.isDirected();
        int nodesNet=net.getNumNodes();
        this.outAdjList=new TIntObjectHashMap[nodesNet];
        this.inAdjList=new TIntObjectHashMap[nodesNet];
        this.nodeListColors=new int[nodesNet];
        this.listEdgeCols=net.getListEdgeCols();
        this.mapEdgeColsIds=net.getMapEdgeColsIds();
        this.maxMultiplicity=net.getMaxMultiplicity();
        this.name=net.getName();
        for(int i=0;i<nodesNet;i++)
        {
            outAdjList[i] = new TIntObjectHashMap<>();
            inAdjList[i] = new TIntObjectHashMap<>();
        }
    }

    public void addNode(int id, int colorNode)
    {
        nodeListColors[id]=colorNode;
        if(colorNode+1>numNodeCols)
            numNodeCols=colorNode+1;
    }

    public void addEdge(int source, int dest, int colEdge)
    {
        if(!outAdjList[source].containsKey(dest))
        {
            TIntHashSet setCols=new TIntHashSet();
            setCols.add(colEdge);
            outAdjList[source].put(dest,setCols);
            inAdjList[dest].put(source,setCols);
            if(!directed && !outAdjList[dest].containsKey(source))
            {
                TIntHashSet setCols2=new TIntHashSet();
                setCols2.add(colEdge);
                outAdjList[dest].put(source,setCols2);
                inAdjList[source].put(dest,setCols2);
            }
        }
        else
        {
            TIntHashSet setColsEdgePart=outAdjList[source].get(dest);
            setColsEdgePart.add(colEdge);
            setColsEdgePart=inAdjList[dest].get(source);
            setColsEdgePart.add(colEdge);
            if(!directed)
            {
                setColsEdgePart=outAdjList[dest].get(source);
                setColsEdgePart.add(colEdge);
                setColsEdgePart=inAdjList[source].get(dest);
                setColsEdgePart.add(colEdge);
            }
        }
    }

    public void addEdge(int source, int dest, TIntHashSet setColsEdge)
    {
        outAdjList[source].put(dest,setColsEdge);
        inAdjList[dest].put(source,setColsEdge);
        if(!directed)
        {
            outAdjList[dest].put(source,setColsEdge);
            inAdjList[source].put(dest,setColsEdge);
        }
    }

    public boolean isEdge(int source, int dest)
    {
        return outAdjList[source].containsKey(dest);
    }

    public int getNumNodes()
    {
        return outAdjList.length;
    }

    void setListEdgeCols()
    {
        int i;
        maxMultiplicity=0;
        for(i=0;i<outAdjList.length;i++)
        {
            TIntIterator it=outAdjList[i].keySet().iterator();
            while(it.hasNext())
            {
                int adjNode=it.next();
                TIntHashSet setEdgeCols=outAdjList[i].get(adjNode);
                if(setEdgeCols.size()>maxMultiplicity)
                    maxMultiplicity=setEdgeCols.size();
                int index=mapEdgeColsIds.size();
                if(!mapEdgeColsIds.containsKey(setEdgeCols))
                {
                    mapEdgeColsIds.put(setEdgeCols,index);
                    listEdgeCols.add(setEdgeCols);
                }
            }
        }
    }

    public Vector<TIntHashSet> getListEdgeCols()
    {
        return listEdgeCols;
    }

    public TObjectIntCustomHashMap<TIntHashSet> getMapEdgeColsIds()
    {
        return mapEdgeColsIds;
    }

    public int[] getNodeListColors()
    {
        return nodeListColors;
    }

    public TIntObjectHashMap[] getOutAdjList()
    {
        return outAdjList;
    }

    public TIntObjectHashMap[] getInAdjList()
    {
        return inAdjList;
    }

    public boolean isDirected()
    {
        return directed;
    }

    TIntDoubleHashMap[][][] getMultiColorDegDistr()
    {
        int i,j,k;
        int numEdgeCols=mapEdgeColsIds.size();
        TIntDoubleHashMap[][][] multiDegDistr=new TIntDoubleHashMap[2][numNodeCols][numEdgeCols];
        for(i=0;i<multiDegDistr.length;i++)
        {
            for(j=0;j<multiDegDistr[i].length;j++)
            {
                for(k=0;k<multiDegDistr[i][j].length;k++)
                    multiDegDistr[i][j][k]=new TIntDoubleHashMap();
            }
        }
        int[] numTotalNodes=new int[numNodeCols];
        for(i=0;i<outAdjList.length;i++)
        {
            int colNode=nodeListColors[i];
            numTotalNodes[colNode]++;
            int[] degDistrNodeOut=new int[numEdgeCols];
            int[] degDistrNodeIn=new int[numEdgeCols];
            TIntIterator it=outAdjList[i].keySet().iterator();
            while(it.hasNext())
            {
                int adjNode=it.next();
                TIntHashSet setEdgeCols=outAdjList[i].get(adjNode);
                int posCol=mapEdgeColsIds.get(setEdgeCols);
                degDistrNodeOut[posCol]++;

            }
            for(k=0;k<degDistrNodeOut.length;k++)
                multiDegDistr[0][colNode][k].adjustOrPutValue(degDistrNodeOut[k],1.0,1.0);
            it=inAdjList[i].keySet().iterator();
            while(it.hasNext())
            {
                int adjNode=it.next();
                TIntHashSet setEdgeCols=inAdjList[i].get(adjNode);
                int posCol=mapEdgeColsIds.get(setEdgeCols);
                degDistrNodeIn[posCol]++;
            }
            for(k=0;k<degDistrNodeIn.length;k++)
                multiDegDistr[1][colNode][k].adjustOrPutValue(degDistrNodeIn[k],1.0,1.0);
        }
        for(i=0;i<multiDegDistr.length;i++)
        {
            for(j=0;j<multiDegDistr[i].length;j++)
            {
                for(k=0;k<multiDegDistr[i][j].length;k++)
                {
                    TIntIterator it=multiDegDistr[i][j][k].keySet().iterator();
                    while(it.hasNext())
                    {
                        int deg=it.next();
                        double valDeg=multiDegDistr[i][j][k].get(deg);
                        if(numTotalNodes[j]!=0)
                        {
                            valDeg = valDeg / numTotalNodes[j];
                                multiDegDistr[i][j][k].put(deg,valDeg);
                        }
                    }
                }
            }
        }
        return multiDegDistr;
    }

    public double[] getFreqNodeColors()
    {
        int numNodes=outAdjList.length;
        double[] freqColors=new double[numNodeCols];
        int i;
        for(i=0;i<nodeListColors.length;i++)
            freqColors[nodeListColors[i]]++;
        for(i=0;i<freqColors.length;i++)
            freqColors[i]/=numNodes;
        return freqColors;
    }

    HashMap<String,TCustomHashSet<TIntHashSet>> getCondEdgeColors()
    {
        HashMap<String,TCustomHashSet<TIntHashSet>> condEdgeCols=new HashMap<>();
        int i;
        for(i=0;i<outAdjList.length;i++)
        {
            int colorSource=nodeListColors[i];
            TIntIterator it=outAdjList[i].keySet().iterator();
            while(it.hasNext())
            {
                int adjNode=it.next();
                int colorDest=nodeListColors[adjNode];
                TIntHashSet setEdgeCols=outAdjList[i].get(adjNode);
                String key=colorSource+"-"+colorDest;
                if(condEdgeCols.containsKey(key))
                {
                    TCustomHashSet<TIntHashSet> setMultiEdges=condEdgeCols.get(key);
                    setMultiEdges.add(setEdgeCols);
                }
                else
                {
                    TCustomHashSet<TIntHashSet> setMultiEdges=new TCustomHashSet<>(new IntHashSetStrategy());
                    setMultiEdges.add(setEdgeCols);
                    condEdgeCols.put(key,setMultiEdges);
                }
            }
        }
        return condEdgeCols;
    }

    HashMap<Integer,TCustomHashSet<TIntHashSet>> getCondEdgeColors2()
    {
        HashMap<Integer,TCustomHashSet<TIntHashSet>> condEdgeCols=new HashMap<>();
        int i;
        for(i=0;i<outAdjList.length;i++)
        {
            int colorSource=nodeListColors[i];
            TIntIterator it=outAdjList[i].keySet().iterator();
            while(it.hasNext())
            {
                int adjNode=it.next();
                TIntHashSet setEdgeCols=outAdjList[i].get(adjNode);
                if(condEdgeCols.containsKey(colorSource))
                {
                    TCustomHashSet<TIntHashSet> setMultiEdges=condEdgeCols.get(colorSource);
                    setMultiEdges.add(setEdgeCols);
                }
                else
                {
                    TCustomHashSet<TIntHashSet> setMultiEdges=new TCustomHashSet<>(new IntHashSetStrategy());
                    setMultiEdges.add(setEdgeCols);
                    condEdgeCols.put(colorSource,setMultiEdges);
                }
            }
        }
        return condEdgeCols;
    }

    public int getNumNodeCols()
    {
        return numNodeCols;
    }

    public int getMaxMultiplicity()
    {
        return maxMultiplicity;
    }

    public String toString()
    {
        String str="";
        int i;
        for(i=0;i<outAdjList.length;i++)
        {
            str+="("+i+","+nodeListColors[i]+") --> ";
            TIntIterator it=outAdjList[i].keySet().iterator();
            while(it.hasNext())
            {
                int dest=it.next();
                str += "(" + dest;
                TIntHashSet setEdgeCols=outAdjList[i].get(dest);
                TIntIterator itCols=setEdgeCols.iterator();
                while(itCols.hasNext())
                {
                    int lab=itCols.next();
                    str+=","+lab;
                }
                str+=") ";
            }
            str+="\n";
        }
        return str;
    }

    public String getName()
    {
        return name;
    }

    public void setName(String newName)
    {
        name=newName;
    }

    void setNumEdgeCols(int numEdgeCols)
    {
        this.numEdgeCols=numEdgeCols;
    }

    int getNumEdgeCols()
    {
        return numEdgeCols;
    }

}
