import java.util.*;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.custom_hash.TObjectIntCustomHashMap;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;

public class GraphUtility
{

	static Vector<Integer>[] getSymmetryConditions(Motif graph)
	{
		int i, j, k;
		Vector<int[]> vv=graph.getAutos();
		int size=graph.getNumNodes();
		Vector<Integer>[] listCond=new Vector[size];
		for(i=0;i<listCond.length;i++)
			listCond[i]=new Vector<>();
		int vvsize=vv.size();
		boolean[] broken=new boolean[vvsize];
		for(i=0;i<size;i++)
		{
			for(j=0;j<vvsize;j++)
			{
				if(!broken[j] && vv.get(j)[i]!=i) 
					break;
			}
			if(j<vvsize)
			{
				for(k=i+1;k<size;k++) 
				{
					for (j=0;j<vvsize;j++)
					{
						if(!broken[j] && vv.get(j)[i]==k) 
						{
							listCond[k].add(i);
							break;
						}
					}
				}
			}
			for(j=0;j<vvsize;j++)
			{
				if(vv.get(j)[i]!=i)
					broken[j]=true;
			}
		}
		return listCond;
	}

	public static Vector<int[]> findAutomorphisms(Motif m)
	{
		int i, j, k;
		int g;
		int[] nodeLabels = m.getNodeLabels();
		int nof_nodes = nodeLabels.length;
		TIntObjectHashMap[] adjLists = m.getOutAdjLists();
		int[] fDir = new int[nof_nodes];
		int[] fRev = new int[nof_nodes];
		for (i = 0; i < nof_nodes; i++) {
			fDir[i] = -1;
			fRev[i] = -1;
		}
		int[][] sequence = new int[nof_nodes][nof_nodes];
		for (i = 0; i < nof_nodes; i++)
		{
			for (j = 0; j < nof_nodes; j++)
			{
				if (adjLists[i].containsKey(j) || adjLists[j].containsKey(i))
				{
					int numNeighs = 0;
					for (k = 0; k < nof_nodes; k++)
					{
						if (adjLists[j].containsKey(k))
							numNeighs++;
						if (adjLists[k].containsKey(j) && !adjLists[j].containsKey(k))
							numNeighs++;
					}
					sequence[i][j] = numNeighs;
				} else {
					sequence[i][j] = 0;
				}
			}
		}
		for (i = 0; i < nof_nodes; i++)
			Arrays.sort(sequence[i]);

		boolean[] support = new boolean[nof_nodes * nof_nodes];
		for (i = 0; i < nof_nodes * nof_nodes; i++)
			support[i] = false;
		for (i = 0; i < nof_nodes; i++)
		{
			for (j = 0; j < nof_nodes; j++)
			{
				for (k = 0; k < nof_nodes; k++)
				{
					if (sequence[i][k] != sequence[j][k])
						break;
				}
				if (k >= nof_nodes)
					support[i * nof_nodes + j] = true;
			}
		}
		Vector<int[]> vv=new Vector<>();
		for (g = 0; g < nof_nodes; g++)
		{
			if (support[g * nof_nodes])
			{
				fDir[0] = g;
				fRev[g] = 0;
				int pos = 1;
				isomorphicExtensions(fDir, fRev, vv, support, pos, nodeLabels, adjLists);
				fRev[fDir[0]] = -1;
				fDir[0] = -1;
			}
		}
		return vv;
	}

	private static void isomorphicExtensions(int[] fDir, int[] fRev, Vector<int[]> vv, boolean[] support, int pos,
									  int[] nodeLabels, TIntObjectHashMap[] adjLists)
	{
		int i, j;
		int nof_nodes = fDir.length;
		int[] cand = new int[nof_nodes];
		int ncand;
		int num;
		for (i = 0; i < nof_nodes; i++)
			cand[i] = -1;
		if (pos == nof_nodes)
		{
			int[] vTemp=new int[nof_nodes];
			for(i=0;i<fDir.length;i++)
				vTemp[i]=fDir[i];
			vv.add(vTemp);
		}
		else
		{
			int n, m;
			boolean flag;
			int[] count = new int[nof_nodes];
			ncand = 0;
			for (i = 0; i < nof_nodes; i++)
				count[i] = 0;
			for (i = 0; i < nof_nodes; i++)
			{
				if (fDir[i] != -1)
				{
					Vector<Integer> vNei = new Vector<>();
					for (j = 0; j < nof_nodes; j++)
					{
						if (adjLists[i].containsKey(j))
							vNei.add(j);
						else if (adjLists[j].containsKey(i))
							vNei.add(j);
					}
					num = vNei.size();
					for (j = 0; j < num; j++)
					{
						int neigh = vNei.get(j);
						if (fDir[neigh] == -1)
						{
							if (count[neigh] == 0)
								cand[ncand++] = neigh;
							count[neigh]++;
						}
					}
				}
			}
			m = 0;
			for (i = 1; i < ncand; i++)
			{
				if (count[i] > count[m])
					m = i;
			}
			m = cand[m];
			ncand = 0;
			boolean[] already = new boolean[nof_nodes];
			for (i = 0; i < nof_nodes; i++)
				already[i] = false;
			for (i = 0; i < nof_nodes; i++)
			{
				if (fDir[i] != -1)
				{
					Vector<Integer> vNei = new Vector<>();
					for (j = 0; j < nof_nodes; j++)
					{
						if (adjLists[fDir[i]].containsKey(j))
							vNei.add(j);
						else if (adjLists[j].containsKey(fDir[i]))
							vNei.add(j);
					}
					num = vNei.size();
					for (j = 0; j < num; j++)
					{
						int neigh = vNei.get(j);
						if (!already[neigh] && fRev[neigh] == -1 && support[m * nof_nodes + neigh])
						{
							cand[ncand++] = neigh;
							already[neigh] = true;
						}
					}
				}
			}
			for (i = 0; i < ncand; i++)
			{
				n = cand[i];
				flag = false;
				if (nodeLabels[n]==nodeLabels[m])
				{
					for (j = 0; j < nof_nodes; j++)
					{
						if (fDir[j] != -1)
						{
							TIntHashSet edgeLabs1=(TIntHashSet) adjLists[m].get(j);
							TIntHashSet edgeLabs2=(TIntHashSet) adjLists[n].get(fDir[j]);
							if ((edgeLabs1==null && edgeLabs2!=null) ||
									(edgeLabs1!=null && edgeLabs2==null) ||
									(edgeLabs1!=null && edgeLabs2!=null && !edgeLabs1.equals(edgeLabs2)))
							{
								flag = true;
								break;
							}
							else
							{
								edgeLabs1=(TIntHashSet) adjLists[j].get(m);
								edgeLabs2=(TIntHashSet) adjLists[fDir[j]].get(n);
								if ((edgeLabs1==null && edgeLabs2!=null) ||
										(edgeLabs1!=null && edgeLabs2==null) ||
										(edgeLabs1!=null && edgeLabs2!=null && !edgeLabs1.equals(edgeLabs2)))
								{
									flag = true;
									break;
								}
							}

						}
					}
					if (!flag)
					{
						fDir[m] = n;
						fRev[n] = m;
						pos++;
						isomorphicExtensions(fDir, fRev, vv, support, pos, nodeLabels, adjLists);
						pos--;
						fRev[fDir[m]] = -1;
						fDir[m] = -1;
					}
				}
			}
		}
	}

	static void expandMotif(Motif m, int maxMultiplicity, int numEdgeCols, HashMap<Integer,TCustomHashSet<TIntHashSet>> condEdgeColors, EDDModel edd, boolean approx)
	{
		int[] nodeLabs=m.getNodeLabels();
		TIntObjectHashMap[] adjLists=m.getOutAdjLists();
		boolean directed=m.isDirected();
		int i, j, k;

		Motif temp=new Motif(directed,nodeLabs.length);
		for(j=0;j<nodeLabs.length;j++)
			temp.addNode(j,nodeLabs[j]);
		for(j=0;j<adjLists.length;j++)
		{
			TIntObjectHashMap<TIntHashSet> adjList=adjLists[j];
			TIntIterator it=adjList.keySet().iterator();
			while(it.hasNext())
			{
				k=it.next();
				if(directed || j<k)
					temp.addEdge(j,k,new TIntHashSet(-1));
			}
		}
		temp.setAutos();
		Vector<int[]> autosTemp=temp.getAutos();

		Vector<Vector<TIntHashSet>> setPossEdges=new Vector<>();
		for(i=0;i<adjLists.length;i++)
		{
			TIntObjectHashMap<TIntHashSet> adjList=adjLists[i];
			TCustomHashSet<TIntHashSet> condEdge=condEdgeColors.get(nodeLabs[i]);
			TIntIterator it=adjList.keySet().iterator();
			while(it.hasNext())
			{
				j=it.next();
				if(directed || i<j)
				{
					Vector<Integer> setPossLabels=new Vector<>();
					TIntHashSet edge=adjList.get(j);
					for(k=0;k<numEdgeCols;k++)
					{
						if(!edge.contains(k))
							setPossLabels.add(k);
					}
					Vector<TIntHashSet> v=new Vector<>();
					if(condEdge.contains(edge))
						v.add(edge);
					MathUtility.getPowerSet(v,edge,condEdge,setPossLabels,maxMultiplicity-edge.size());
					setPossEdges.add(v);
				}
			}
		}
		MathUtility.getCombinations(setPossEdges, 0, new TIntHashSet[setPossEdges.size()], edd, m, autosTemp, approx);
	}

	static Motif createSuperMotif(Motif mSource, Motif mDest, int s)
	{
		int[] setLabsSource=mSource.getNodeLabels();
		int[] setLabsDest=mDest.getNodeLabels();
		TIntObjectHashMap[] adjSource=mSource.getOutAdjLists();
		TIntObjectHashMap[] adjDest=mDest.getOutAdjLists();
		boolean directed=mSource.isDirected();
		int k=adjSource.length;

		int i;
		Motif mSuper=new Motif(directed,2*k-s);

		//Add nodes to super motif
		for(i=0;i<setLabsSource.length;i++)
		{
			mSuper.addNode(i,setLabsSource[i]);
			mSuper.addNode(i+k-s,setLabsDest[i]);
		}

		//Add edges to super motif
		int l, m;
		//Copy adjSource in the super-motif adjMatrix
		for(l=0;l<adjSource.length;l++)
		{
			TIntIterator it=adjSource[l].keySet().iterator();
			while(it.hasNext())
			{
				m=it.next();
				mSuper.addEdge(l,m,(TIntHashSet) adjSource[l].get(m));
			}
		}
		//Copy adjDest in the super-motif adjMatrix
		for(l=0;l<adjDest.length;l++)
		{
			TIntIterator it=adjDest[l].keySet().iterator();
			while(it.hasNext())
			{
				m=it.next();
				mSuper.addEdge(l+k-s,m+k-s,(TIntHashSet) adjDest[l].get(m));
			}
		}
		//Compute logical OR in the middle part of super-motif adjMatrix
		TIntObjectHashMap[] adjSuper=mSuper.getOutAdjLists();
		for(l=k-s;l<k;l++)
		{
			for (m=k-s;m<k;m++)
			{
				TIntHashSet edgeSuper=(TIntHashSet) adjSuper[l].get(m);
				if(edgeSuper==null)
				{
					TIntHashSet edgeSource=(TIntHashSet) adjSource[l].get(m);
					if(edgeSource!=null)
						mSuper.addEdge(l,m,edgeSource);
				}
			}
		}

		return mSuper;
	}

	static TCustomHashSet<Motif> getNonRedPerm(Motif mot)
	{
		TCustomHashSet<Motif> nonRedPerm=new TCustomHashSet<>(new MotifStrategy());
		int[] nodeColsOcc=mot.getNodeLabels();
		TIntObjectHashMap[] edgeColsOcc=mot.getOutAdjLists();
		boolean directed=mot.isDirected();
		int[] currentPerm=new int[nodeColsOcc.length];
		int i, j;
		for(i=0; i<currentPerm.length;i++)
			currentPerm[i]=i;
		//Add the indentity permutation to the set of non redundant permutations
		nonRedPerm.add(mot);
		//Compute all possible permutations and add the corrisponding motif to the non redundant list, if not present
		for(;MathUtility.nextPermutation(currentPerm);)
		{
			Motif newMot=new Motif(directed,nodeColsOcc.length);
			//Get the new set of colors according to the new permutation
			for(i=0;i<currentPerm.length;i++)
				newMot.addNode(currentPerm[i],nodeColsOcc[i]);
			//Get the adjacency matrix corresponding to the new permutation
			for(i=0;i<currentPerm.length;i++)
			{
				TIntIterator it=edgeColsOcc[i].keySet().iterator();
				while(it.hasNext())
				{
					j=it.next();
					TIntHashSet edge=(TIntHashSet) edgeColsOcc[i].get(j);
					newMot.addEdge(currentPerm[i],currentPerm[j],edge);
				}
			}
			nonRedPerm.add(newMot);
		}
		return nonRedPerm;
	}

	static LinkedList<Integer>[][][] getListDegrees(Graph net)
	{
		int i,j,k;

		int[] colorNodeNet=net.getNodeListColors();
		int numNodeCols=net.getNumNodeCols();
		TObjectIntCustomHashMap<TIntHashSet> mapEdgeColsIds=net.getMapEdgeColsIds();
		int numEdgeCols=mapEdgeColsIds.size();

		//Assign node color-dependent degrees
		LinkedList<Integer>[][][] listDegrees=new LinkedList[2][numNodeCols][numEdgeCols];
		for(i=0;i<numNodeCols;i++)
		{
			for(j=0;j<numEdgeCols;j++)
			{
				listDegrees[0][i][j]=new LinkedList<>();
				listDegrees[1][i][j]=new LinkedList<>();
			}
		}
		TIntObjectHashMap[] netOutAdjList=net.getOutAdjList();
		TIntObjectHashMap[] netInAdjList=net.getInAdjList();
		int numNodes=netOutAdjList.length;
		for(i=0;i<numNodes;i++)
		{
			int colNode=colorNodeNet[i];
			int[] degreesOut=new int[numEdgeCols];
			int[] degreesIn=new int[numEdgeCols];
			TIntIterator it=netOutAdjList[i].keySet().iterator();
			while(it.hasNext())
			{
				int adjNode=it.next();
				TIntHashSet setEdgeCols=(TIntHashSet) netOutAdjList[i].get(adjNode);
				int posCol=mapEdgeColsIds.get(setEdgeCols);
				degreesOut[posCol]++;
			}
			it=netInAdjList[i].keySet().iterator();
			while(it.hasNext())
			{
				int adjNode=it.next();
				TIntHashSet setEdgeCols=(TIntHashSet) netInAdjList[i].get(adjNode);
				int posCol=mapEdgeColsIds.get(setEdgeCols);
				degreesIn[posCol]++;
			}
			for(k=0;k<numEdgeCols;k++)
			{
				listDegrees[0][colNode][k].add(degreesOut[k]);
				listDegrees[1][colNode][k].add(degreesIn[k]);
			}
		}
		return listDegrees;
	}

	static LinkedList<Integer>[][] assignDegrees(int[] nodeLabelsRand, LinkedList<Integer>[][][] listDegrees, boolean directed)
	{
		int i, j, k, l;
		Random r=new Random();
		int numEdgeCols=listDegrees[0][0].length;
		LinkedList<Integer>[][] degWeights=new LinkedList[2][numEdgeCols];
		for(i=0;i<degWeights.length;i++)
		{
			for(j=0;j<numEdgeCols;j++)
				degWeights[i][j]=new LinkedList<>();
		}
		int numNodes=nodeLabelsRand.length;
		for(j=0;j<numNodes;j++)
		{
			int colNode=nodeLabelsRand[j];
			for(k=0;k<numEdgeCols;k++)
			{
				int outDeg=listDegrees[0][colNode][k].get(r.nextInt(listDegrees[0][colNode][k].size()));
				int inDeg;
				if(directed)
					inDeg=listDegrees[1][colNode][k].get(r.nextInt(listDegrees[1][colNode][k].size()));
				else
					inDeg=outDeg;
				for(l=0;l<outDeg;l++)
					degWeights[0][k].add(j);
				for(l=0;l<inDeg;l++)
					degWeights[1][k].add(j);
			}
		}
		return degWeights;
	}

}