import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TIntHashSet;

public class MatchingMachine
{
	public enum NodeFlag {NS_CORE, NS_CNEIGH, NS_UNV}
	public int nof_sn;
	public int[] edges_sizes;				//indexed by state_id
	public MaMaEdge[][] edges;				//indexed by state_id, map on states  (0,1) = (state0, state1)
	public int[] map_node_to_state;			//indexed by node_id
	public int[] map_state_to_node;			//indexed by state_id
	public int[] parent_state;				//indexed by state_id
	public MamaParentType[] parent_type;	//indexed by state id
	
	public MatchingMachine(TIntObjectHashMap[] queryOutAdjList, TIntObjectHashMap[] queryInAdjList)
	{
		nof_sn = queryOutAdjList.length;
		edges_sizes = new int[nof_sn];
		edges = new MaMaEdge[nof_sn][];
		map_node_to_state = new int[nof_sn];
		map_state_to_node = new int[nof_sn];
		parent_state = new int[nof_sn];
		parent_type = new MamaParentType[nof_sn];
		build(queryOutAdjList,queryInAdjList);
	}
	
	public void build(TIntObjectHashMap[] outAdiacs, TIntObjectHashMap[] inAdiacs)
	{
		int i, j;
		NodeFlag[] node_flags = new NodeFlag[nof_sn]; //indexed by node_id
		int[][] weights = new int[nof_sn][3]; //indexed by node_id
		int[] t_parent_node = new int[nof_sn]; //indexed by node_id
		MamaParentType[] t_parent_type = new MamaParentType[nof_sn]; //indexed by node id
		for(i=0; i<nof_sn; i++)
		{
			node_flags[i] = NodeFlag.NS_UNV;
			weights[i] = new int[3];
			weights[i][0] = 0;
			weights[i][1] = 0;
			//weights[i][2] = ssg.out_adj_sizes[i] + ssg.in_adj_sizes[i];
			weights[i][2] = outAdiacs[i].size() + inAdiacs[i].size();
			t_parent_node[i] = -1;
			t_parent_type[i] = MamaParentType.PARENTTYPE_NULL;
		}
		int si = 0;
		int n;
		int nIT; int ni;
		int nni;
		int nqueueL = 0, nqueueR = 0;
		int maxi, maxv;
		int tmp;
		while(si < nof_sn)
		{
			if(nqueueL == nqueueR)
			{
				//if queue is empty....
				maxi = -1;
				maxv = -1;
				nIT = 0;
				while(nIT < nof_sn)
				{
					if(node_flags[nIT]==NodeFlag.NS_UNV &&  weights[nIT][2] > maxv)
					{
						maxv = weights[nIT][2];
						maxi = nIT;
					}
					nIT++;
				}
				map_state_to_node[si] = maxi;
				map_node_to_state[maxi] = si;
				t_parent_type[maxi] = MamaParentType.PARENTTYPE_NULL;
				t_parent_node[maxi] = -1;
				nqueueR++;
				n = maxi;
				TIntIterator it=outAdiacs[n].keySet().iterator();
				while(it.hasNext())
				{
					ni=it.next();
					if(ni != n)
						weights[ni][1]++;
				}
				it=inAdiacs[n].keySet().iterator();
				while(it.hasNext())
				{
					ni=it.next();
					if(ni != n)
						weights[ni][1]++;
				}
			}
			if(nqueueL != nqueueR-1)
			{
				maxi = nqueueL;
				for(int mi=maxi+1; mi<nqueueR; mi++)
				{
					if(wcompare(map_state_to_node[mi], map_state_to_node[maxi], weights) < 0)
						maxi = mi;
				}
				tmp = map_state_to_node[nqueueL];
				map_state_to_node[nqueueL] = map_state_to_node[maxi];
				map_state_to_node[maxi] = tmp;
			}
			n = map_state_to_node[si];
			map_node_to_state[n] = si;
			//move queue left limit
			nqueueL++;
			//update nodes' flags & weights
			node_flags[n] = NodeFlag.NS_CORE;
			TIntIterator it=outAdiacs[n].keySet().iterator();
			while(it.hasNext())
			{
				ni=it.next();
				if(ni != n)
				{
					weights[ni][0]++;
					weights[ni][1]--;
					if(node_flags[ni] == NodeFlag.NS_UNV)
					{
						node_flags[ni] = NodeFlag.NS_CNEIGH;
						t_parent_node[ni] = n;
//						if(nIT < ssg.out_adj_sizes[n])
							t_parent_type[ni] = MamaParentType.PARENTTYPE_OUT;
//						else
//							t_parent_type[ni] = PARENTTYPE_IN;
						//add to queue
						map_state_to_node[nqueueR] = ni;
						map_node_to_state[ni] = nqueueR;
						nqueueR++;
						TIntIterator it2=outAdiacs[ni].keySet().iterator();
						while(it2.hasNext())
						{
							nni=it2.next();
							weights[nni][1]++;
						}
					}
				}
			}
			it=inAdiacs[n].keySet().iterator();
			while(it.hasNext())
			{
				ni=it.next();
				if(ni != n)
				{
					weights[ni][0]++;
					weights[ni][1]--;
					if(node_flags[ni] == NodeFlag.NS_UNV)
					{
						node_flags[ni] = NodeFlag.NS_CNEIGH;
						t_parent_node[ni] = n;
//						if(nIT < ssg.out_adj_sizes[n])
//							t_parent_type[ni] = PARENTTYPE_OUT;
//						else
							t_parent_type[ni] = MamaParentType.PARENTTYPE_IN;
						//add to queue
						map_state_to_node[nqueueR] = ni;
						map_node_to_state[ni] = nqueueR;
						nqueueR++;
						TIntIterator it2=inAdiacs[n].keySet().iterator();
						while(it2.hasNext())
						{
							nni=it2.next();
							weights[nni][1]++;
						}
					}
				}
			}
			si++;
		}
		int e_count;
		for(si = 0; si<nof_sn; si++)
		{
			n = map_state_to_node[si];
			if(t_parent_node[n] != -1)
				parent_state[si] = map_node_to_state[t_parent_node[n]];
			else
				parent_state[si] = -1;
			parent_type[si] = t_parent_type[n];
			e_count = 0;
			TIntIterator it=outAdiacs[n].keySet().iterator();
			while(it.hasNext())
			{
				int idOut=it.next();
				if(map_node_to_state[idOut]<si)
					e_count++;
			}
			it=inAdiacs[n].keySet().iterator();
			while(it.hasNext())
			{
				int idIn=it.next();
				if(map_node_to_state[idIn]<si)
					e_count++;
			}
			edges_sizes[si] = e_count;
			edges[si] = new MaMaEdge[e_count];
			e_count = 0;
			it=outAdiacs[n].keySet().iterator();
			while(it.hasNext())
			{
				int idOut=it.next();
				if(map_node_to_state[idOut] < si)
				{
					TIntHashSet setCols=(TIntHashSet) outAdiacs[n].get(idOut);
					edges[si][e_count]=new MaMaEdge(map_node_to_state[n],map_node_to_state[idOut],setCols);
					//edges[si][e_count].source = map_node_to_state[n];
					//edges[si][e_count].target = map_node_to_state[idOut];
					e_count++;
				}
			}
			for(j=0; j<si; j++)
			{
				int sn = map_state_to_node[j];
				it=outAdiacs[sn].keySet().iterator();
				while(it.hasNext())
				{
					int idOut=it.next();
					if(idOut==n)
					{
						TIntHashSet setCols=(TIntHashSet) outAdiacs[sn].get(idOut);
						edges[si][e_count]=new MaMaEdge(j,si,setCols);
						//edges[si][e_count].source = j;
						//edges[si][e_count].target = si;
						e_count++;
					}
				}
			}
		}
	}
	
	private int wcompare(int i, int j, int[][] weights)
	{
		for(int w=0; w<3; w++)
		{
			if(weights[i][w] != weights[j][w])
				return weights[j][w] - weights[i][w];
		}
		return i-j;
	}

}
