import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.Vector;

public class MathUtility
{
	private static int matBinomCoeff[][]={ 
	{1,  0,  0,   0,   0,   0,   0,   0,  0,  0, 0} ,
	{1,  1,  0,   0,   0,   0,   0,   0,  0,  0, 0} ,
	{1,  2,  1,   0,   0,   0,   0,   0,  0,  0, 0} , 
	{1,  3,  3,   1,   0,   0,   0,   0,  0,  0, 0} , 
	{1,  4,  6,   4,   1,   0,   0,   0,  0,  0, 0} , 
	{1,  5, 10,  10,   5,   1,   0,   0,  0,  0, 0} ,
	{1,  6, 15,  20,  15,   6,   1,   0,  0,  0, 0} ,
	{1,  7, 21,  35,  35,  21,   7,   1,  0,  0, 0} ,
	{1,  8, 28,  56,  70,  56,  28,   8,  1,  0, 0} ,
	{1,  9, 36,  84, 126, 126,  84,  36,  9,  1, 0} ,
	{1, 10, 45, 120, 210, 252, 210, 120, 45, 10, 1} 
	};

	static double binomCoeff(int n, int k)
	{
		if(k>n)
			return 0;
		else 
		{
			if(k==0)
				return 1;
			else if(k==1)
				return n; 
			else if(n<=10) {
				return matBinomCoeff[n][k];
			} 
			else 
			{
				//Vandermonde identity
				if(k>(n/2))
					k=n-k;
				int m1=n/2;
				int m2=n-m1;
				double res=0;
				if(m1!=m2) 
					// Pascal triangle
					res=binomCoeff(2*m1,k)+binomCoeff(2*m1,k-1);
				else 
				{
					int k_end ;
					if(k%2==1) 
						//Odd
						k_end=(k+1)/2;
					else 
					{
						//even
						k_end=k/2;
						res=binomCoeff(m1,k_end );
						res=res*res;
					}
					for(int i=0;i<k_end;i++)
						res+=2*(binomCoeff(m1,i)*binomCoeff(m1,k-i));
				}
				return res;
			}
		}
	}

	static double multinomCoeff(int n,int[] k)
	{
		int max=-1;
		int i_max=-1;
		for(int i=0;i<k.length;i++) 
		{
			if(k[i]>max ) 
			{
				i_max=i;
				max=k[i];
			}
		}
		double res=factorial(n,max);
		double denom=1;
		for(int i=0;i<k.length;i++) 
		{
			if(i!=i_max)
				denom=denom*factorial(k[i],1); 
		}
		res=res/denom;
		return res;
	}

	private static double factorial(int n,int k)
	{
		double res = (n>=(k+1))? (k + 1) : 1 ; 
		for(int i=k+2;i<=n;i++) 
			res*= i;
		return res;
	}

	private static double factorial(int n)
	{
		double res = 1.0;
		for(int i=2;i<=n;i++)
			res*= i;
		return res;
	}
	
	static boolean nextPermutation(int[] p)
	{
		for (int a = p.length - 2; a >= 0; --a)
		{
			if (p[a] < p[a + 1])
			{
				for (int b = p.length - 1;; --b)
				{
					if (p[b] > p[a]) 
					{
						int t = p[a];
						p[a] = p[b];
						p[b] = t;
						for (++a, b = p.length - 1; a < b; ++a, --b) 
						{
							t = p[a];
							p[a] = p[b];
							p[b] = t;
						}
						return true;
					}
				}
			}
		}
		return false;
	}

	public static Vector<TIntHashSet> getPowerSet(TIntHashSet setColors)
	{
		Vector<TIntHashSet> powerSet=new Vector<>();
		int n = setColors.size();
		for (int i = 0; i < (1<<n); i++)
		{
			TIntHashSet set=new TIntHashSet();
			TIntIterator it=setColors.iterator();
			for (int j = 0; j < n; j++)
			{
				int col=it.next();
				if ((i & (1 << j)) > 0)
					set.add(col);
			}
			if(set.size()>0)
				powerSet.add(set);
		}
		return powerSet;
	}

	static void getCombinations(Vector<Vector<TIntHashSet>> sets, int n, TIntHashSet[] temp, EDDModel edd, Motif mot, Vector<int[]> autosTemp, boolean approx)
	{
		if(n>=sets.size())
		{
			int i;
			TIntHashSet[] copyTemp=new TIntHashSet[temp.length];
			for(i=0;i<temp.length;i++)
				copyTemp[i]=temp[i];

			int[] nodeLabs=mot.getNodeLabels();
			TIntObjectHashMap[] adjLists=mot.getOutAdjLists();
			boolean directed=mot.isDirected();
			Motif extMot=new Motif(directed,nodeLabs.length);
			for(i=0;i<nodeLabs.length;i++)
				extMot.addNode(i,nodeLabs[i]);
			int l=0;
			for(i=0;i<adjLists.length;i++)
			{
				TIntObjectHashMap<TIntHashSet> adjList=adjLists[i];
				TIntIterator it=adjList.keySet().iterator();
				while(it.hasNext())
				{
					int j=it.next();
					if(directed || i<j)
					{
						extMot.addEdge(i,j,copyTemp[l]);
						l++;
					}
				}
			}
			extMot=extMot.getCanonicalForm(autosTemp);

			if(approx)
			{
				int motifSize=mot.getNumNodes();
				RISolverKocay riKocay=new RISolverKocay(extMot);
				riKocay.solve(mot);
				extMot.setAutos();
				double numAutos = extMot.getAutos().size();
				double numNRP = MathUtility.factorial(motifSize) / numAutos;
				double meanExt=edd.computeMean(extMot, numNRP);
				mot.setAvgAnal(mot.getAvgAnal()+meanExt*riKocay.getNumOccurrences());
			}
			else
			{
				RISolverKocay riKocay=new RISolverKocay(extMot);
				riKocay.solve(mot);
				TCustomHashSet<Motif> nonRedPerms=GraphUtility.getNonRedPerm(extMot);
				double meanExt=edd.computeMean(extMot,nonRedPerms.size());
				mot.setAvgAnal(mot.getAvgAnal()+meanExt*riKocay.getNumOccurrences());
				//System.out.println("Mean anal: "+meanAnal);
				double varianceExt=edd.computeVariance(extMot,nonRedPerms,meanExt);
				mot.setVarianceAnal(mot.getVarianceAnal()+varianceExt*riKocay.getNumOccurrences());
				//System.out.println("Variance anal: "+varianceAnal);
				//System.out.println("Std dev anal: "+Math.sqrt(varianceAnal));
			}

			return;
		}
		int i;
		Vector<TIntHashSet> set=sets.get(n);
		for(i=0;i<set.size();i++)
		{
			TIntHashSet s=set.get(i);
			temp[n]=s;
			getCombinations(sets, n+1, temp, edd, mot, autosTemp, approx);
		}
	}

	public static Vector<TIntHashSet> getPowerSet(Vector<Integer> setLabels, int minSize, int maxSize)
	{
		Vector<TIntHashSet> powerSet=new Vector<>();
		int n = setLabels.size();
		for (int i = 0; i < (1<<n); i++)
		{
			TIntHashSet set=new TIntHashSet();
			for (int j = 0; j < n; j++)
			{
				int lab=setLabels.get(j);
				if ((i & (1 << j)) > 0)
					set.add(lab);
			}
			if(set.size()>=minSize && set.size()<=maxSize)
				powerSet.add(set);
		}
		return powerSet;
	}

	public static void getPowerSet(Vector<TIntHashSet> v, TIntHashSet edge, TCustomHashSet<TIntHashSet> condEdge, Vector<Integer> setLabels, int maxSize)
	{
		int n = setLabels.size();
		for (int i = 0; i < (1<<n); i++)
		{
			TIntHashSet set=new TIntHashSet();
			for (int j = 0; j < n; j++)
			{
				int lab=setLabels.get(j);
				if ((i & (1 << j)) > 0)
					set.add(lab);
			}
			if(set.size()>0 && set.size()<=maxSize)
			{
				TIntHashSet newComb=new TIntHashSet(edge);
				newComb.addAll(set);
				if(condEdge.contains(newComb))
					v.add(newComb);
			}
		}
	}

	static int[] sortCondAuto(int[] nodeLabels, Vector<int[]> autos)
	{
		int i, j;
		int[] bestAuto=new int[nodeLabels.length];
		for(i=0;i<bestAuto.length;i++)
			bestAuto[i]=autos.get(0)[i];
		for(i=1;i<autos.size();i++)
		{
			int[] auto=autos.get(i);
			boolean minusThan=false;
			for(j=0;j<auto.length;j++)
			{
				//System.out.println(colors[auto[j]]+"-"+colors[bestAuto[j]]);
				if(nodeLabels[auto[j]]<nodeLabels[bestAuto[j]])
				{
					minusThan=true;
					break;
				}
				else if(nodeLabels[auto[j]]>nodeLabels[bestAuto[j]])
					break;
			}
			if(minusThan)
			{
				for(j=0;j<auto.length;j++)
					bestAuto[j]=auto[j];
			}
		}
		int[] sortedLabels=new int[nodeLabels.length];
		for(i=0;i<bestAuto.length;i++)
			sortedLabels[i]=nodeLabels[bestAuto[i]];
		return sortedLabels;
	}

	static void getNodeCombs(TCustomHashSet<int[]> setCombs, Vector<Integer>[] sets, int n, int[] temp, Vector<int[]> autosDummy)
	{
		if(n==sets.length)
		{
			int i;
			int[] copyTemp=new int[temp.length];
			for(i=0;i<temp.length;i++)
				copyTemp[i]=temp[i];
			copyTemp=sortCondAuto(copyTemp,autosDummy);
			setCombs.add(copyTemp);
			return;
		}
		int i;
		for(i=0;i<sets[n].size();i++)
		{
			int s=sets[n].get(i);
			temp[n]=s;
			getNodeCombs(setCombs, sets, n+1, temp, autosDummy);
		}
	}

	static void getEdgeCombs(Vector<TIntHashSet[]> setCombs, Vector<Vector<TIntHashSet>> sets, int n, TIntHashSet[] temp)
	{
		if(n==sets.size())
		{
			int i;
			TIntHashSet[] copyTemp=new TIntHashSet[temp.length];
			for(i=0;i<temp.length;i++)
				copyTemp[i]=temp[i];
			setCombs.add(copyTemp);
			return;
		}
		int i;
		Vector<TIntHashSet> set=sets.get(n);
		for(i=0;i<set.size();i++)
		{
			TIntHashSet s=set.get(i);
			temp[n]=s;
			getEdgeCombs(setCombs, sets, n+1, temp);
		}
	}
}
	