import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TIntHashSet;

import java.util.Vector;

public class Motif
{
    private boolean directed;
	private int[] nodeLabels;
    private TIntObjectHashMap<TIntHashSet>[] outAdjLists;
	private TIntObjectHashMap<TIntHashSet>[] inAdjLists;
    private long numOccurrences;
    private double pvalSimul;
    private double avgRandom;
    private double pvalAnal;
    private double avgAnal;
    private double varianceAnal;
    private Vector<int[]> setAutos;
    private Vector<Integer>[] symmCond;
    private MatchingMachine matMachine;
	
	public Motif(boolean directed, int numNodes)
	{
		this.directed=directed;
		this.nodeLabels=new int[numNodes];
		this.outAdjLists=new TIntObjectHashMap[numNodes];
		this.inAdjLists=new TIntObjectHashMap[numNodes];
		int i;
		for(i=0;i<numNodes;i++)
		{
			outAdjLists[i]=new TIntObjectHashMap<>();
			inAdjLists[i]=new TIntObjectHashMap<>();
		}
		this.numOccurrences=0L;
        this.pvalSimul=0.0;
        this.avgRandom=0.0;
        this.pvalAnal=0.0;
        this.avgAnal=0.0;
        this.varianceAnal=0.0;
	}
	
	public void addNode(int id, int nodeLabel)
    {
        nodeLabels[id]=nodeLabel;
    }

    public void addEdge(int source, int dest, TIntHashSet edgeLabels)
    {
        outAdjLists[source].put(dest,edgeLabels);
		inAdjLists[dest].put(source,edgeLabels);
        if(!directed)
        {
            outAdjLists[dest].put(source,edgeLabels);
			inAdjLists[source].put(dest,edgeLabels);
        }
    }
	
    public int[] getNodeLabels()
    {
        return nodeLabels;
    }
	
    public int getNumNodes()
    {
        return nodeLabels.length;
    }
	
    public TIntObjectHashMap[] getOutAdjLists()
    {
        return outAdjLists;
    }
	
    TIntObjectHashMap[] getInAdjLists()
    {
        return inAdjLists;
    }
	
    public boolean isDirected()
    {
	    return directed;
    }
    
    public long getNumOccurrences()
    {
        return numOccurrences;
    }
	
    public void setNumOccurrences(long numOccs)
    {
        numOccurrences=numOccs;
    }
	
    public double getPValSimul()
    {
        return pvalSimul;
    }
	
    void setPValSimul(double val)
    {
        pvalSimul=val;
    }
	
    public double getAvgRandom()
    {
        return avgRandom;
    }
	
    void setAvgRandom(double val)
    {
        avgRandom=val;
    }
	
    double getPValAnal()
    {
        return pvalAnal;
    }
	
    void setPValAnal(double val)
    {
        pvalAnal=val;
    }
	
    double getAvgAnal()
    {
        return avgAnal;
    }
	
    void setAvgAnal(double val)
    {
        avgAnal=val;
    }
	
    double getVarianceAnal()
    {
        return varianceAnal;
    }
	
    void setVarianceAnal(double val)
    {
        varianceAnal=val;
    }
	
    Vector<int[]> getAutos()
    {
        return setAutos;
    }
	
    void setAutos()
    {
        Vector<int[]> autos=GraphUtility.findAutomorphisms(this);
        setAutos=autos;
    }
	
    Vector<Integer>[] getSymmCond()
    {
        return symmCond;
    }
	
    void setSymmCond()
    {
        Vector<Integer>[] symmConditions=GraphUtility.getSymmetryConditions(this);
        symmCond=symmConditions;
    }
	
    void setMatMachine()
    {
        matMachine=new MatchingMachine(outAdjLists,inAdjLists);
    }
	
    MatchingMachine getMatMachine()
    {
        return matMachine;
    }
	
    public boolean isEquivalentTo(Motif occ)
    {
        int[] refNodeLabels=occ.getNodeLabels();
        TIntObjectHashMap[] refOutAdjLists=occ.getOutAdjLists();
        boolean[] matched=new boolean[refNodeLabels.length];
        int i, j;
        for(i=0;i<nodeLabels.length;i++)
        {
            for(j=0;j<refNodeLabels.length;j++)
            {
                if(!matched[j] && nodeLabels[i]==refNodeLabels[j])
                {
                    boolean[] matchedEdges=new boolean[refNodeLabels.length];
					TIntIterator itSource=outAdjLists[i].keySet().iterator();
					boolean checkOut=true;
					while(itSource.hasNext())
                    {
						int k=itSource.next();
						TIntIterator it=outAdjLists[i].keySet().iterator();
						boolean checkIn=false;
						while(it.hasNext())
						{
                            int l=it.next();
							if(!matchedEdges[l] && nodeLabels[k]==refNodeLabels[l] && (outAdjLists[i].get(k)).equals(refOutAdjLists[j].get(l)))
                            {
                                matchedEdges[l]=true;
								checkIn=true;
                                break;
                            }
                        }
                        if(!checkIn)
						{
                            checkOut=false;
							break;
						}
                    }
                    if(checkOut)
                    {
                        matched[j] = true;
                        //System.out.println(i+" --> "+j);
                        break;
                    }
                }
            }
            if(j==refNodeLabels.length)
                return false;
        }
        return i==nodeLabels.length;
    }

    Motif getCanonicalForm(Vector<int[]> autos)
    {
        int i, j, k;
        //System.out.println(autos.size());
        int[] bestAuto=new int[nodeLabels.length];
        for(i=0;i<bestAuto.length;i++)
            bestAuto[i]=autos.get(0)[i];
        for(i=1;i<autos.size();i++)
        {
            int[] auto=autos.get(i);
            boolean minusThan=false;
            for(j=0;j<auto.length;j++)
            {
                //System.out.println(colors[auto[j]]+"-"+colors[bestAuto[j]]);
                if(nodeLabels[auto[j]]<nodeLabels[bestAuto[j]])
                {
                    minusThan=true;
                    break;
                }
                else if(nodeLabels[auto[j]]>nodeLabels[bestAuto[j]])
                    break;
                else
                {
                    boolean complete=false;
                    for(k=0;k<auto.length;k++)
                    {
                        TIntHashSet edge1=outAdjLists[auto[j]].get(auto[k]);
                        if(edge1!=null)
                        {
                            TIntHashSet edge2=outAdjLists[bestAuto[j]].get(bestAuto[k]);
                            //System.out.println(edge1+" - "+edge2);
                            String str1=edge1.toString();
                            String str2=edge2.toString();
                            if(str1.compareTo(str2)<0)
                            {
                                minusThan=true;
                                complete=true;
                                break;
                            }
                            else if(str1.compareTo(str2)>0)
                            {
                                complete=true;
                                break;
                            }
                        }
                    }
                    if(complete)
                        break;
                }
            }
            if(minusThan)
            {
                for(j=0;j<auto.length;j++)
                    bestAuto[j]=auto[j];
            }
        }
        Motif canon=new Motif(directed,nodeLabels.length);
        for(i=0;i<bestAuto.length;i++)
            canon.addNode(i,nodeLabels[bestAuto[i]]);
        for(i=0;i<bestAuto.length;i++)
        {
            for(j=0;j<bestAuto.length;j++)
            {
                if(outAdjLists[bestAuto[i]].containsKey(bestAuto[j]))
                    canon.addEdge(i,j,outAdjLists[bestAuto[i]].get(bestAuto[j]));
            }
        }
        //System.out.println(canon);
        return canon;
    }

    public String toString()
    {
        int i;
        String str="";
        for(i=0;i<nodeLabels.length-1;i++)
            str+=nodeLabels[i]+",";
        str+=nodeLabels[i]+" ----- ";
        for(i=0;i<outAdjLists.length;i++)
        {
            TIntObjectHashMap<TIntHashSet> adjList=outAdjLists[i];
			TIntIterator it=adjList.keySet().iterator();
			while(it.hasNext())
            {
                int j=it.next();
				if(directed || i<j)
                {
                    str+="("+i+","+j;
					TIntHashSet setEdges=adjList.get(j);
                    TIntIterator itEdge=setEdges.iterator();
                    while(itEdge.hasNext())
                    {
                        int edgeLab=itEdge.next();
                        str+=","+edgeLab;
                    }
                    str+=")";
                }
            }
        }
        return str;
    }
	
}
