import gnu.trove.map.hash.TCustomHashMap;
import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

public class MotifManager
{
    private Graph target;
    private TCustomHashSet<Query> queries;

    public MotifManager(Graph target, TCustomHashSet<Query> queries)
    {
        this.target=target;
        this.queries=queries;
    }

    public TCustomHashSet<Motif> censusGraph(int minFreq, HashMap<String,TCustomHashSet<TIntHashSet>> condEdgeColors)
    {
        RISolver solver=new RISolver(target);
        System.out.println("Estimating motif frequencies in input network... 0%");
        double status=0.1;
        int i=0;
        Iterator<Query> it=queries.iterator();
        int numQueries=queries.size();
        TCustomHashSet<Motif> setFinalMotifs=new TCustomHashSet<>(new MotifStrategy());
        while(it.hasNext())
        {
            Query query=it.next();
            solver.solve(query);
            TCustomHashMap<Motif,Long> mapOccs=solver.getMapOccurrences();
            if(!mapOccs.isEmpty())
            {
                TCustomHashMap<int[],TCustomHashMap<Motif,Long>> newMapOccs=new TCustomHashMap<>(new IntArrayStrategy());
                Vector<int[]> autosDummy=query.getDummyAutos();

                for(Motif m : mapOccs.keySet())
                {
                    long numOccs=mapOccs.get(m);
                    Motif canon=m.getCanonicalForm(autosDummy);
                    int[] nodeLabels=canon.getNodeLabels();
                    nodeLabels=MathUtility.sortCondAuto(nodeLabels,autosDummy);

                    if(!newMapOccs.containsKey(nodeLabels))
                    {
                        TCustomHashMap<Motif,Long> newMapLabeledOccs=new TCustomHashMap<>(new MotifStrategy());
                        canon.setAutos();
                        int numAutos=canon.getAutos().size();
                        newMapLabeledOccs.put(canon,numOccs/numAutos);
                        newMapOccs.put(nodeLabels,newMapLabeledOccs);
                    }
                    else
                    {
                        TCustomHashMap<Motif,Long> newMapLabeledOccs=newMapOccs.get(nodeLabels);
                        if(!newMapLabeledOccs.containsKey(canon))
                        {
                            canon.setAutos();
                            int numAutos=canon.getAutos().size();
                            newMapLabeledOccs.put(canon,numOccs/numAutos);
                        }
                    }
                }

                int[] queryNodeLabels=query.getNodeLabels();
                Vector<Integer>[] combNodeLabs=new Vector[queryNodeLabels.length];
                int k, j;
                int numNodeLabs=target.getNumNodeCols();
                int numEdgeLabs=target.getNumEdgeCols();
                for(k=0;k<queryNodeLabels.length;k++)
                {
                    combNodeLabs[k]=new Vector<>();
                    if(queryNodeLabels[k]==-1)
                    {
                        for(j=0;j<numNodeLabs;j++)
                            combNodeLabs[k].add(j);
                    }
                    else
                        combNodeLabs[k].add(queryNodeLabels[k]);
                }
                TCustomHashSet<int[]> setNodeCombs=new TCustomHashSet<>(new IntArrayStrategy());
                MathUtility.getNodeCombs(setNodeCombs, combNodeLabs, 0, new int[combNodeLabs.length],autosDummy);

                for(int[] nodeComb : setNodeCombs)
                {
                    TCustomHashSet<Motif> setInst=query.getInstantiationSet(nodeComb,condEdgeColors,numEdgeLabs,autosDummy);
                    for(Motif g: setInst)
                    {
                        g.setAutos();
                        g.setSymmCond();
                        g.setMatMachine();
                        long numInstOccs=0;
                        TCustomHashMap<Motif,Long> newMapLabeledOccs=newMapOccs.get(nodeComb);
                        if(newMapLabeledOccs!=null)
                        {
                            for(Motif graphDest : newMapLabeledOccs.keySet())
                            {
                                RISolverKocay riComb=new RISolverKocay(graphDest);
                                riComb.solve(g);
                                numInstOccs+=riComb.getNumOccurrences()*newMapLabeledOccs.get(graphDest);
                            }
                            if(numInstOccs>=minFreq)
                            {
                                g.setNumOccurrences(numInstOccs);
                                setFinalMotifs.add(g);
                            }

                        }
                    }
                }

                //TCustomHashSet<Motif> setInstantiations=GraphUtility.instantiateQuery(query,condEdgeColors,numNodeLabs,target.getNumEdgeCols(),setMotifs,minFreq);
                //setFinalMotifs.addAll(setInstantiations);
                //System.out.println(setInstantiations.size());
            }
            if(((double)i)/numQueries>=status)
            {
                status+=0.1;
                System.out.println("Estimating motif frequencies in input network... "+(int)(((double)i+1)/numQueries*100)+"%");
            }
            i++;
        }
        return setFinalMotifs;
    }

    void censusRandom(EDDModel edd, int numRandVars, TCustomHashSet<Motif> setFinalMotifs, double pvalSimulThresh, String randomFolder, FileManager fm)
    {
        int i;

        LinkedList<Integer>[][][] listDegrees=null;
        if(randomFolder==null)
            listDegrees=GraphUtility.getListDegrees(target);

        System.out.println("Estimating motif frequencies in EDD random graphs... 0%");
        double status=0.1;
        for(i=0;i<numRandVars;i++)
        {
            Graph randNet;
            if(randomFolder==null)
                randNet=edd.sampleGraph(target,listDegrees);
            else
                randNet=fm.readGraph(randomFolder+"/"+target.getName()+"_"+(i+1)+".txt");
            RISolver solver=new RISolver(randNet);
            for(Query query : queries)
            {
                solver.solve(query);
                TCustomHashMap<Motif,Long> mapOccs=solver.getMapOccurrences();

                if(!mapOccs.isEmpty())
                {
                    TCustomHashMap<int[],TCustomHashMap<Motif,Long>> newMapOccs=new TCustomHashMap<>(new IntArrayStrategy());
                    Vector<int[]> autosDummy=query.getDummyAutos();

                    for(Motif m : mapOccs.keySet())
                    {
                        long numOccs=mapOccs.get(m);
                        Motif canon=m.getCanonicalForm(autosDummy);
                        int[] nodeLabels=canon.getNodeLabels();
                        nodeLabels=MathUtility.sortCondAuto(nodeLabels,autosDummy);

                        if(!newMapOccs.containsKey(nodeLabels))
                        {
                            TCustomHashMap<Motif,Long> newMapLabeledOccs=new TCustomHashMap<>(new MotifStrategy());
                            canon.setAutos();
                            int numAutos=canon.getAutos().size();
                            newMapLabeledOccs.put(canon,numOccs/numAutos);
                            newMapOccs.put(nodeLabels,newMapLabeledOccs);
                        }
                        else
                        {
                            TCustomHashMap<Motif,Long> newMapLabeledOccs=newMapOccs.get(nodeLabels);
                            if(!newMapLabeledOccs.containsKey(canon))
                            {
                                canon.setAutos();
                                int numAutos=canon.getAutos().size();
                                newMapLabeledOccs.put(canon,numOccs/numAutos);
                            }
                        }
                    }

                    for(Motif motSource : setFinalMotifs)
                    {
                        long numFinalOccsRand=0;
                        int[] nodeLabs=motSource.getNodeLabels();
                        nodeLabs=MathUtility.sortCondAuto(nodeLabs,autosDummy);

                        TCustomHashMap<Motif,Long> newMapLabeledOccs=newMapOccs.get(nodeLabs);
                        if(newMapLabeledOccs!=null)
                        {
                            for(Motif graphDest : newMapLabeledOccs.keySet())
                            {
                                RISolverKocay riComb=new RISolverKocay(graphDest);
                                riComb.solve(motSource);
                                numFinalOccsRand+=riComb.getNumOccurrences()*newMapLabeledOccs.get(graphDest);
                            }
                            long numOccsInput=motSource.getNumOccurrences();
                            if(numFinalOccsRand>=numOccsInput)
                                motSource.setPValSimul(motSource.getPValSimul()+1);
                            motSource.setAvgRandom(motSource.getAvgRandom()+numFinalOccsRand);
                        }

                        /*for(Motif motDest : setMotifs)
                        {
                            RISolverKocay riKocay=new RISolverKocay(motDest);
                            riKocay.solve(motSource);
                            numFinalOccsRand+=riKocay.getNumOccurrences()*motDest.getNumOccurrences();
                        }
                        long numOccsInput=motSource.getNumOccurrences();
                        if(numFinalOccsRand>=numOccsInput)
                            motSource.setPValSimul(motSource.getPValSimul()+1);
                        motSource.setAvgRandom(motSource.getAvgRandom()+numFinalOccsRand);*/
                    }

                }
            }
            if(((double)i)/numRandVars>=status)
            {
                System.out.println("Estimating motif frequencies in EDD random graphs... "+(int)(Math.round(status*100))+"%");
                status+=0.1;
            }
        }
        if(numRandVars>0)
        {
            Iterator<Motif> it=setFinalMotifs.iterator();
            while(it.hasNext())
            {
                Motif mot=it.next();
                double pval=mot.getPValSimul()/numRandVars;
                if(pval<=pvalSimulThresh)
                {
                    mot.setPValSimul(pval);
                    mot.setAvgRandom(mot.getAvgRandom()/numRandVars);
                }
                else
                    it.remove();
            }
        }
    }

    void computeAnalyticalPVals(EDDModel edd, TCustomHashSet<Motif> setMotifs, double pvalAnalThresh, boolean approx)
    {
        int i=0;
        System.out.println("Computing analytical p-values... 0%");
        double status=0.1;
        int maxMultiplicity=target.getMaxMultiplicity();
        int numEdgeCols=target.getNumEdgeCols();
        HashMap<Integer,TCustomHashSet<TIntHashSet>> condEdgeColors=target.getCondEdgeColors2();
        Iterator<Motif> it=setMotifs.iterator();
        while(it.hasNext())
        {
            Motif mot=it.next();
            long numOccs=mot.getNumOccurrences();

            PolyaAeppli pa=new PolyaAeppli();
            double pvalAnal;
            GraphUtility.expandMotif(mot,maxMultiplicity,numEdgeCols,condEdgeColors,edd,approx);

            if(approx)
                pvalAnal = pa.estimateMaxPvalue(numOccs, mot.getAvgAnal());
            else
                pvalAnal=pa.computePValue(numOccs,mot.getAvgAnal(),mot.getVarianceAnal());

            if(pvalAnal<-1.0E-17)
                pvalAnal=0.0;
            else if(pvalAnal>1.0)
                pvalAnal=1.0;
            else if(Double.isNaN(pvalAnal))
            {
                if(mot.getAvgAnal()<1.0)
                    pvalAnal=0.0;
                else
                    pvalAnal=1.0;
            }
            if(pvalAnal<=pvalAnalThresh)
                mot.setPValAnal(pvalAnal);
            else
                it.remove();
            if(((double)i)/setMotifs.size()>=status)
            {
                status+=0.1;
                System.out.println("Computing analytical p-values... "+(int)(((double)i+1)/setMotifs.size()*100)+"%");
            }
            i++;
        }
    }

}
