import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TIntHashSet;
import gnu.trove.strategy.HashingStrategy;

public class MotifStrategy implements HashingStrategy<Motif>
{
    public int computeHashCode(Motif o)
    {
        int h = 31; // seed chosen at random
        int i;
        int[] nodeLabs=o.getNodeLabels();
		TIntObjectHashMap[] outAdjLists=o.getOutAdjLists();
        for(i=0;i<nodeLabs.length;i++)
            h = h ^ ((h << 5) + (h >> 2) + nodeLabs[i]);
        for(i=0;i<outAdjLists.length;i++)
        {
            TIntObjectHashMap<TIntHashSet> adjList=outAdjLists[i];
			TIntIterator it=adjList.keySet().iterator();
			while(it.hasNext())
			{
                int j=it.next();
				h = h ^ ((h << 5) + (h >> 2) + (adjList.get(j)).hashCode());
            }
        }
        return h;
    }
    public boolean equals(Motif o1, Motif o2)
    {
        int i, j;
        int[] nodeLabs1=o1.getNodeLabels();
        int[] nodeLabs2=o2.getNodeLabels();
        for(i=0;i<nodeLabs1.length;i++)
        {
            if(nodeLabs1[i]!=nodeLabs2[i])
                return false;
        }
        TIntObjectHashMap[] adjListsOut1=o1.getOutAdjLists();
        TIntObjectHashMap[] adjListsOut2=o2.getOutAdjLists();
        TIntObjectHashMap[] adjListsIn1=o1.getInAdjLists();
        TIntObjectHashMap[] adjListsIn2=o2.getInAdjLists();
        for(i=0;i<adjListsOut1.length;i++)
        {
			for(j=0;j<adjListsOut1.length;j++)
            {
                TIntHashSet edge1=(TIntHashSet) adjListsOut1[i].get(j);
                TIntHashSet edge2=(TIntHashSet) adjListsOut2[i].get(j);
                if((edge1==null && edge2!=null) ||
                (edge1!=null && edge2==null) || (edge1!=null && edge2!=null && !edge1.equals(edge2)))
                    return false;
                if(o1.isDirected())
                {
                    edge1=(TIntHashSet) adjListsIn1[i].get(j);
                    edge2=(TIntHashSet) adjListsIn2[i].get(j);
                    if((edge1==null && edge2!=null) ||
                            (edge1!=null && edge2==null) || (edge1!=null && edge2!=null && !edge1.equals(edge2)))
                        return false;
                }
            }
        }
        return true;
    }
}
