import gnu.trove.set.hash.TCustomHashSet;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYTitleAnnotation;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTickUnit;
import org.jfree.chart.block.BlockBorder;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.title.LegendTitle;
import org.jfree.chart.title.TextTitle;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.RectangleAnchor;
import org.jfree.ui.RectangleEdge;

import java.awt.Color;
import java.util.Vector;

class PlotManager
{
	private String plotTitle;
	private String plotSubtitle;
	private String netName;
	PlotManager(String netFile, int motifSize)
	{
		String[] split=netFile.split("[\\\\|/]");
		String[] split2=split[split.length-1].split("\\.");
		netName=split2[0];
		//netName=netFile;
		plotTitle="Precision-Recall plot for "+netName;
		plotSubtitle="Motifs of size "+motifSize+" with color-topology dependence";
	}
	
	JFreeChart createPRPlot(TCustomHashSet<Motif> setMotifs, double thresholdPVal)
	{
		Vector<String> precRecValues=getPrecRecallValues(setMotifs,thresholdPVal);
		XYDataset dataset=createDataset(netName,precRecValues);
		JFreeChart chart=createChart(plotTitle,plotSubtitle,dataset);
		return chart;
	}
	
	private static XYDataset createDataset(String netName, Vector<String> precRecValues)
	{
		XYSeries series=new XYSeries(netName);
		int i;
		for(i=0;i<precRecValues.size();i++)
		{
			String[] split=precRecValues.get(i).split("-");
			double recall=Double.parseDouble(split[0]);
			double prec=Double.parseDouble(split[1]);
			series.add(recall,prec);
		}
		XYSeriesCollection dataset=new XYSeriesCollection();
		dataset.addSeries(series);
		return dataset;
	}
	
	private static JFreeChart createChart(String titleText, String subtitleText, XYDataset dataset)
	{
		JFreeChart chart=ChartFactory.createXYLineChart(
		titleText,
		"Recall (%)",
		"Precision (%)",
		dataset,
		PlotOrientation.VERTICAL,
		false, //include legend
		true, //tooltips
		false //urls
		);
		//Change plot features
		XYPlot plot=(XYPlot) chart.getPlot();
		plot.setBackgroundPaint(Color.white);
		//Change range axis
		NumberAxis domain = (NumberAxis) plot.getDomainAxis();
		domain.setRange(0, 102);
		domain.setTickUnit(new NumberTickUnit(5));
		NumberAxis range = (NumberAxis) plot.getRangeAxis();
		range.setRange(0,102);
		range.setTickUnit(new NumberTickUnit(10));
		//Place legend inside plot
		LegendTitle lt = new LegendTitle(plot);
		lt.setFrame(new BlockBorder(Color.black));
		lt.setPosition(RectangleEdge.BOTTOM);
		XYTitleAnnotation ta = new XYTitleAnnotation(0.02, 0.02, lt,RectangleAnchor.BOTTOM_LEFT);
		ta.setMaxWidth(0.2);
		plot.addAnnotation(ta);
		//Add a subtitle for the chart
		TextTitle subtitle=new TextTitle(subtitleText);
		chart.addSubtitle(subtitle);
		return chart;
	}

	private static Vector<String> getPrecRecallValues(TCustomHashSet<Motif> setMotifs, double thresholdPval)
	{
		Vector<Double> pvalsAnal=new Vector<>();
		Vector<Double> pvalsSimul=new Vector<>();
		int i;
		for(Motif mot : setMotifs)
		{
			double pvalAnal=mot.getPValAnal();
			double pvalSimul=mot.getPValSimul();
			pvalsAnal.add(pvalAnal);
			pvalsSimul.add(pvalSimul);
		}
		double pvalRef;
		String lastValPrecRec="";
		Vector<String> precRecList=new Vector<>();
		for(pvalRef=0.0;pvalRef<=1.0;pvalRef+=0.0001)
		{
			double numTruePositives=0.0;
			double numFalsePositives=0.0;
			double numFalseNegatives=0.0;
			for(i=0;i<pvalsSimul.size();i++)
			{
				double simulPval=pvalsSimul.get(i);
				double analPval=pvalsAnal.get(i);
				if(simulPval<=thresholdPval)
				{
					if(analPval<=pvalRef)
						numTruePositives++;
					else
						numFalseNegatives++;
				}
				else
				{
					if(analPval<=pvalRef)
						numFalsePositives++;
				}
			}
			double precision=1.0;
			double recall=1.0;
			if(numTruePositives+numFalsePositives>0.0)
				precision=numTruePositives/(numTruePositives+numFalsePositives);
			if(numTruePositives+numFalseNegatives>0.0)
				recall=numTruePositives/(numTruePositives+numFalseNegatives);
			precision*=100;
			recall*=100;

			String valPrecRec=recall+"-"+precision;
			if(!valPrecRec.equals(lastValPrecRec))
			{
				precRecList.add(valPrecRec);
				lastValPrecRec=valPrecRec;
			}

		}
		return precRecList;
	}
}