import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;
import gnu.trove.map.hash.TIntIntHashMap;

import java.util.HashMap;
import java.util.Vector;

public class Query
{
    private boolean directed;
	private int[] nodeLabels;
    private TIntObjectHashMap<TIntHashSet>[] outAdjLists;
    private TIntObjectHashMap<TIntHashSet>[] inAdjLists;
    private TIntIntHashMap[] minEdgeMulti;
    private TIntIntHashMap[] maxEdgeMulti;
    private MatchingMachine matMachine;
	public Query(int numNodes, boolean directed)
	{
		this.directed=directed;
		this.nodeLabels=new int[numNodes];
        this.outAdjLists=new TIntObjectHashMap[numNodes];
        this.inAdjLists=new TIntObjectHashMap[numNodes];
        this.minEdgeMulti=new TIntIntHashMap[numNodes];
        this.maxEdgeMulti=new TIntIntHashMap[numNodes];
		int i;
		for(i=0;i<numNodes;i++)
		{
            outAdjLists[i]=new TIntObjectHashMap<>();
            inAdjLists[i]=new TIntObjectHashMap<>();
            minEdgeMulti[i]=new TIntIntHashMap();
            maxEdgeMulti[i]=new TIntIntHashMap();
		}
	}
	
	public void addNode(int id, int nodeLabel)
    {
        nodeLabels[id]=nodeLabel;
    }

    public void addEdge(int source, int dest, TIntHashSet edgeLabels, int minMulti, int maxMulti)
    {
        outAdjLists[source].put(dest,edgeLabels);
        inAdjLists[dest].put(source,edgeLabels);
        minEdgeMulti[source].put(dest,minMulti);
        maxEdgeMulti[source].put(dest,maxMulti);
        if(!directed)
        {
            outAdjLists[dest].put(source,edgeLabels);
            inAdjLists[source].put(dest,edgeLabels);
            minEdgeMulti[dest].put(source,minMulti);
            maxEdgeMulti[dest].put(source,maxMulti);
        }
    }
	
    public int[] getNodeLabels()
    {
        return nodeLabels;
    }
	
    public int getNumNodes()
    {
        return nodeLabels.length;
    }

    public TIntObjectHashMap[] getOutAdjLists()
    {
        return outAdjLists;
    }

    TIntObjectHashMap[] getInAdjLists()
    {
        return inAdjLists;
    }

    TIntIntHashMap[] getMaxEdgeMulti()
    {
        return maxEdgeMulti;
    }

    void setMatMachine()
    {
        matMachine=new MatchingMachine(outAdjLists,inAdjLists);
    }

    MatchingMachine getMatMachine()
    {
        return matMachine;
    }

	public boolean isDirected()
	{
		return directed;
	}

    Vector<int[]> getDummyAutos()
    {
        int j, k;
        Motif temp=new Motif(directed,nodeLabels.length);
        for(j=0;j<nodeLabels.length;j++)
            temp.addNode(j,-1);
        for(j=0;j<nodeLabels.length;j++)
        {
            TIntIterator it=outAdjLists[j].keySet().iterator();
            while(it.hasNext())
            {
                k=it.next();
                if(directed || j<k)
                    temp.addEdge(j,k,new TIntHashSet(-1));
            }
        }
        return GraphUtility.findAutomorphisms(temp);
    }

    TCustomHashSet<Motif> getInstantiationSet(int[] nodeComb, HashMap<String,TCustomHashSet<TIntHashSet>> condEdgeLabels, int numEdgeCols, Vector<int[]> autosDummy)
    {
        int i, j, k, l;
        TCustomHashSet<Motif> setInst=new TCustomHashSet<>(new MotifStrategy());

        Vector<Vector<TIntHashSet>> combEdgeLabs=new Vector<>();
        for(i=0;i<outAdjLists.length;i++)
        {
            TIntIterator it=outAdjLists[i].keySet().iterator();
            while(it.hasNext())
            {
                j=it.next();
                if(directed || i<j)
                {
                    Vector<TIntHashSet> edgeComb=new Vector<>();
                    TIntHashSet currEdges=outAdjLists[i].get(j);
                    int maxMulti=maxEdgeMulti[i].get(j);
                    int minMulti=minEdgeMulti[i].get(j);
                    Vector<Integer> possEdgeLabs=new Vector<>();
                    for(k=0;k<numEdgeCols;k++)
                    {
                        if(!currEdges.contains(k))
                            possEdgeLabs.add(k);
                    }
                    int maxPossMulti=maxMulti-currEdges.size();
                    int minPossMulti=minMulti-currEdges.size();
                    if(maxPossMulti==0)
                    {
                        String key=nodeComb[i]+"-"+nodeComb[j];
                        if(condEdgeLabels.containsKey(key))
                        {
                            TCustomHashSet<TIntHashSet> setPossEdges=condEdgeLabels.get(key);
                            if(setPossEdges.contains(currEdges))
                                edgeComb.add(currEdges);
                        }
                    }
                    else
                    {
                        Vector<TIntHashSet> possEdgeSets=MathUtility.getPowerSet(possEdgeLabs,minPossMulti,maxPossMulti);
                        String key=nodeComb[i]+"-"+nodeComb[j];
                        if(condEdgeLabels.containsKey(key))
                        {
                            TCustomHashSet<TIntHashSet> setPossEdges=condEdgeLabels.get(key);
                            if(setPossEdges.contains(currEdges))
                                edgeComb.add(currEdges);
                        }
                        for(k=0;k<possEdgeSets.size();k++)
                        {
                            TIntHashSet newComb=new TIntHashSet();
                            newComb.addAll(currEdges);
                            newComb.addAll(possEdgeSets.get(k));
                            key=nodeComb[i]+"-"+nodeComb[j];
                            if(condEdgeLabels.containsKey(key))
                            {
                                TCustomHashSet<TIntHashSet> setPossEdges=condEdgeLabels.get(key);
                                if(setPossEdges.contains(newComb))
                                    edgeComb.add(newComb);
                            }
                        }
                    }
                    combEdgeLabs.add(edgeComb);
                }
            }
        }
        Vector<TIntHashSet[]> setEdgeCombs=new Vector<>();
        MathUtility.getEdgeCombs(setEdgeCombs, combEdgeLabs, 0, new TIntHashSet[combEdgeLabs.size()]);
        for(j=0;j<setEdgeCombs.size();j++)
        {
            TIntHashSet[] edgeComb=setEdgeCombs.get(j);
            Motif inst=new Motif(directed,nodeComb.length);
            for(k=0;k<nodeComb.length;k++)
                inst.addNode(k,nodeComb[k]);
            int x=0;
            for(k=0;k<outAdjLists.length;k++)
            {
                TIntIterator it=outAdjLists[k].keySet().iterator();
                while(it.hasNext())
                {
                    l=it.next();
                    if(directed || k<l)
                    {
                        inst.addEdge(k,l,edgeComb[x]);
                        x++;
                    }
                }
            }
            Motif canon=inst.getCanonicalForm(autosDummy);
            setInst.add(canon);
        }
        return setInst;
    }

    public String toString()
    {
        int i, k;
        String str="";
        for(i=0;i<nodeLabels.length-1;i++)
            str+=nodeLabels[i]+",";
        str+=nodeLabels[i]+" ----- ";
        for(i=0;i<outAdjLists.length;i++)
        {
            TIntObjectHashMap<TIntHashSet> adjList=outAdjLists[i];
			TIntIterator it=adjList.keySet().iterator();
			while(it.hasNext())
            {
                int j=it.next();
                int minMulti=minEdgeMulti[i].get(j);
				int maxMulti=maxEdgeMulti[i].get(j);
				if(directed || i<j)
                {
                    str+="("+i+","+j;
					TIntHashSet setEdges=adjList.get(j);
                    TIntIterator itEdge=setEdges.iterator();
                    while(itEdge.hasNext())
                    {
                        int edgeLab=itEdge.next();
                        str+=","+edgeLab;
                    }
					for(k=setEdges.size();k<minMulti;k++)
						str+=",?";
                    if(k<maxMulti)
                        str+=",*";
                    str+=")";
                }
            }
        }
        return str;
    }
}
