import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TCustomHashMap;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TIntHashSet;

import java.util.HashSet;

public class RISolver
{
	private MatchingMachine mama;
	private Graph targetGraph;
	private Query queryGraph;
	private TCustomHashMap<Motif,Long> mapOccurrences;
	
	public RISolver(Graph targetGraph)
	{
		this.targetGraph=targetGraph;
	}
	
	public void solve(Query queryGraph)
	{
		mapOccurrences=new TCustomHashMap<>(new MotifStrategy());
		this.mama=queryGraph.getMatMachine();
		this.queryGraph=queryGraph;

		boolean directed=queryGraph.isDirected();
		TIntObjectHashMap[] targetOutAdjLists=targetGraph.getOutAdjList();
		TIntObjectHashMap[] targetInAdjLists=targetGraph.getInAdjList();
		TIntObjectHashMap[] queryOutAdjLists=queryGraph.getOutAdjLists();
		int[] targetColors=targetGraph.getNodeListColors();
		int nofTargetNodes=targetOutAdjLists.length;
		int i;
		int nof_sn=mama.nof_sn;
		int[] map_state_to_node=mama.map_state_to_node;
		int[] parent_state=mama.parent_state;
		MamaParentType[] parent_type=mama.parent_type;
		TIntObjectHashMap listAllRef = new TIntObjectHashMap();
		for(i=0;i<nofTargetNodes;i++)
			listAllRef.put(i,null);
		TIntObjectHashMap[] candidates = new TIntObjectHashMap[nof_sn];
		int[] solution = new int[nof_sn];
		for(i=0;i<nof_sn;i++)
			solution[i] = -1;
		HashSet<Integer>[] cmatched=new HashSet[nof_sn];
		for(i=0;i<cmatched.length;i++)
			cmatched[i]=new HashSet<>();
		boolean[] matched = new boolean[nofTargetNodes];
		int[] matchedColors=new int[nof_sn];
		candidates[0] = listAllRef;
		int psi = -1;
		int si = 0;
		int ci;
		int sip1;
		while(si != -1)
		{
			if(psi >= si)
			{
				matched[solution[si]] = false;
				matchedColors[map_state_to_node[si]]=0;
			}
			ci = -1;
			TIntIterator it=candidates[si].keySet().iterator();
			while(it.hasNext())
			{
				ci = it.next();
				solution[si] = ci;
				if(!matched[ci] && !cmatched[si].contains(ci) && nodeCheck(si,ci)
				&& edgesCheck(si,solution))
					break;
				else
					ci=-1;
			}
			if(ci == -1)
			{
				psi = si;
				cmatched[si].clear();
				si--;
			}
			else
			{
				cmatched[si].add(ci);
				matchedColors[map_state_to_node[si]]=targetColors[ci];
				if(si == nof_sn -1)
				{
					Motif occ=new Motif(directed,matchedColors.length);
					for(i=0;i<matchedColors.length;i++)
						occ.addNode(i,matchedColors[i]);
					for(i=0;i<queryOutAdjLists.length;i++)
					{
						int source = solution[mama.map_node_to_state[i]];
						TIntIterator itQuery=queryOutAdjLists[i].keySet().iterator();
						while(itQuery.hasNext())
						{
							int targetQuery=itQuery.next();
							int target = solution[mama.map_node_to_state[targetQuery]];
							TIntHashSet edgeCols = (TIntHashSet) targetOutAdjLists[source].get(target);
							occ.addEdge(i,targetQuery,edgeCols);
						}
					}
					Long numOccs = mapOccurrences.get(occ);
					if(numOccs==null)
						mapOccurrences.put(occ,1L);
					else
						mapOccurrences.put(occ,numOccs+1);
					psi = si;
				}
				else
				{
					matched[solution[si]] = true;
					sip1 = si+1;
					if(parent_type[sip1] == MamaParentType.PARENTTYPE_NULL)
						candidates[sip1] = listAllRef;
					else
					{
						if(parent_type[sip1] == MamaParentType.PARENTTYPE_IN)
							candidates[sip1] = targetInAdjLists[solution[parent_state[sip1]]];
						else
							candidates[sip1] = targetOutAdjLists[solution[parent_state[sip1]]];
					}
					psi = si;
					si++;
				}
			}
		}
	}
	
	protected boolean nodeCheck(int si, int ci)
	{
		boolean nodeCheck=false;
		TIntObjectHashMap[] targetOutAdjLists=targetGraph.getOutAdjList();
		TIntObjectHashMap[] targetInAdjLists=targetGraph.getInAdjList();
		TIntObjectHashMap[] queryOutAdjLists=queryGraph.getOutAdjLists();
		TIntObjectHashMap[] queryInAdjLists=queryGraph.getInAdjLists();
		int[] targetColors=targetGraph.getNodeListColors();
		int[] queryColors=queryGraph.getNodeLabels();
		if((queryColors[mama.map_state_to_node[si]]==-1 || targetColors[ci]==queryColors[mama.map_state_to_node[si]]) &&
		targetOutAdjLists[ci].size()>=queryOutAdjLists[mama.map_state_to_node[si]].size() &&
		targetInAdjLists[ci].size()>=queryInAdjLists[mama.map_state_to_node[si]].size())
			nodeCheck=true;
		return nodeCheck;
	}
	
	private boolean edgesCheck(int si, int[] solution)
	{
		TIntObjectHashMap[] targetOutAdjLists=targetGraph.getOutAdjList();
		for(int me=0; me<mama.edges_sizes[si]; me++)
		{
			int source = solution[mama.edges[si][me].source];
			int target = solution[mama.edges[si][me].target];
			if(!targetGraph.isEdge(source,target))
				return false;
			else
			{
				TIntHashSet setColsQuery=mama.edges[si][me].attr;
				if(setColsQuery!=null)
				{
					TIntHashSet setColsTarget = (TIntHashSet) targetOutAdjLists[source].get(target);
					if(!setColsQuery.isEmpty() && !setColsTarget.containsAll(setColsQuery))
						return false;
				}
			}
		}
		return true;
	}
    TCustomHashMap<Motif, Long> getMapOccurrences()
    {
        return mapOccurrences;
    }
}