import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TIntHashSet;

import java.util.HashSet;
import java.util.Vector;

public class RISolverKocay
{
	private MatchingMachine mama;
	private Motif targetGraph;
	private Motif queryGraph;
	private Vector<Integer>[] symmCond;
	private long numOccurrences;
	
	RISolverKocay(Motif targetGraph)
	{
		this.targetGraph=targetGraph;
	}
	
	public void solve(Motif queryGraph)
	{
		numOccurrences=0;
		this.symmCond=queryGraph.getSymmCond();
		this.mama=queryGraph.getMatMachine();

		this.queryGraph=queryGraph;

		TIntObjectHashMap[] targetOutAdjLists=targetGraph.getOutAdjLists();
		TIntObjectHashMap[] targetInAdjLists=targetGraph.getInAdjLists();
		int nofTargetNodes=targetOutAdjLists.length;
		int i;
		int nof_sn=mama.nof_sn;
		int[] parent_state=mama.parent_state;
		MamaParentType[] parent_type=mama.parent_type;
		TIntObjectHashMap listAllRef = new TIntObjectHashMap();
		for(i=0;i<nofTargetNodes;i++)
			listAllRef.put(i,null);
		TIntObjectHashMap[] candidates = new TIntObjectHashMap[nof_sn];
		int[] solution = new int[nof_sn];
		for(i=0;i<nof_sn;i++)
			solution[i] = -1;
		HashSet<Integer>[] cmatched=new HashSet[nof_sn];
		for(i=0;i<cmatched.length;i++)
			cmatched[i]=new HashSet<>();
		boolean[] matched = new boolean[nofTargetNodes];
		candidates[0] = listAllRef;
		int psi = -1;
		int si = 0;
		int ci;
		int sip1;
		while(si != -1)
		{
			if(psi >= si)
				matched[solution[si]] = false;
			ci = -1;
			TIntIterator it=candidates[si].keySet().iterator();
			while(it.hasNext())
			{
				ci = it.next();
				solution[si] = ci;
				if(!matched[ci] && !cmatched[si].contains(ci) && nodeCheck(si,ci)
				&& condCheck(si,solution) && edgesCheck(si,solution))
					break;
				else
					ci=-1;
			}
			if(ci == -1)
			{
				psi = si;
				cmatched[si].clear();
				si--;
			}
			else
			{
				cmatched[si].add(ci);
				if(si == nof_sn -1)
					numOccurrences++;
				else
				{
					matched[solution[si]] = true;
					sip1 = si+1;
					if(parent_type[sip1] == MamaParentType.PARENTTYPE_NULL)
						candidates[sip1] = listAllRef;
					else
					{
						if(parent_type[sip1] == MamaParentType.PARENTTYPE_IN)
							candidates[sip1] = targetInAdjLists[solution[parent_state[sip1]]];
						else
							candidates[sip1] = targetOutAdjLists[solution[parent_state[sip1]]];
					}
					psi = si;
					si++;
				}
			}
		}
	}
	
	private boolean condCheck(int si, int[] solution)
	{
		boolean condCheck=true;
		//System.out.println(map_state_to_node[si]);
		Vector<Integer> condNode=symmCond[mama.map_state_to_node[si]];
		int ii;
		for(ii=0;ii<condNode.size();ii++)
		{
			int targetNode=solution[mama.map_node_to_state[condNode.get(ii)]];
			if(solution[si]<targetNode)
			{
				condCheck = false;
				break;
			}
		}
		return condCheck;
	}
	
	private boolean nodeCheck(int si, int ci)
	{
		boolean nodeCheck=false;
		TIntObjectHashMap[] targetOutAdjLists=targetGraph.getOutAdjLists();
		TIntObjectHashMap[] targetInAdjLists=targetGraph.getInAdjLists();
		TIntObjectHashMap[] queryOutAdjLists=queryGraph.getOutAdjLists();
		TIntObjectHashMap[] queryInAdjLists=queryGraph.getInAdjLists();
		int[] targetColors=targetGraph.getNodeLabels();
		int[] queryColors=queryGraph.getNodeLabels();
		if(targetColors[ci]==queryColors[mama.map_state_to_node[si]] &&
		targetOutAdjLists[ci].size()>=queryOutAdjLists[mama.map_state_to_node[si]].size() &&
		targetInAdjLists[ci].size()>=queryInAdjLists[mama.map_state_to_node[si]].size())
			nodeCheck=true;
		return nodeCheck;
	}
	
	private boolean edgesCheck(int si, int[] solution)
	{
		TIntObjectHashMap[] targetOutAdjLists=targetGraph.getOutAdjLists();
		for(int me=0; me<mama.edges_sizes[si]; me++)
		{
			int source = solution[mama.edges[si][me].source];
			int target = solution[mama.edges[si][me].target];
			if(!targetOutAdjLists[source].containsKey(target))
				return false;
			else
			{
				TIntHashSet setColsQuery=mama.edges[si][me].attr;
				if(setColsQuery!=null)
				{
					TIntHashSet setColsTarget = (TIntHashSet) targetOutAdjLists[source].get(target);
					if(!setColsTarget.containsAll(setColsQuery))
						return false;
				}
			}
		}
		return true;
	}
    public long getNumOccurrences()
    {
        return numOccurrences;
    }
}