import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;
import org.jfree.chart.JFreeChart;

import java.util.HashMap;

public class TestSimulAnal
{
    public static void main(String[] args)
    {
        //Input parameters
        String netFile="Networks/ppiyeast.txt";
        //String netFile="Networks/internet.txt";
        //String netFile="Networks/openflights.txt";
        //String netFile="netGradoop.txt";
        //String netFile="Networks/imdbMovies.txt";
        //String netFile="Networks/artnetdir.txt";
        String queryFile="query.txt";
        //String queryFile="Queries/ppiyeast/ppiyeast_clique_4_122.txt";
        //String queryFile="Queries/query_3_undir.txt";
        int numRandVars=100;
        int minFreq=1;
        double pvalSimulThresh=1.0;
        double pvalAnalThresh=1.0;
        String outputFile=null;
        //String outputFile="results.txt";
        String outputFilePlot=null;
        //String outputFilePlot="prPlot.pdf";
        //String randomFile="Random/ppiyeast";
        String randomFile=null;
        boolean approx=false;

        long inizio=System.currentTimeMillis();

        //Read input network
        System.out.println("Reading graph file...");
        FileManager fm=new FileManager();
        Graph net=fm.readGraph(netFile);
        boolean directed=net.isDirected();
        int maxMultiplicity=net.getMaxMultiplicity();
        HashMap<String,TCustomHashSet<TIntHashSet>> condEdgeColors=net.getCondEdgeColors();
        //System.out.println(net);

        //Read queries
        System.out.println("Building motif set...");
        TCustomHashSet<Query> setQueries=fm.readQueries(queryFile,directed,maxMultiplicity);

        System.out.println("-----------------------");
        System.out.println("Searching for colored non-induced injective motifs with color-topology dependence...");


        //Build symmetry conditions
        int maxMotifSize=0;
        for(Query q : setQueries)
        {
            q.setMatMachine();
            int motSize=q.getNumNodes();
            if(motSize>maxMotifSize)
                maxMotifSize=motSize;
        }
        MotifManager mm=new MotifManager(net,setQueries);

        //Compute motif frequencies
        TCustomHashSet<Motif> setMotifs=mm.censusGraph(minFreq,condEdgeColors);

        if(maxMotifSize>0)
        {
            System.out.println("Setting up EDD model...");
            EDDModel edd=new EDDModel(net,maxMotifSize,approx);
            //Compute motif frequencies in a sample of EDD random networks
            mm.censusRandom(edd,numRandVars,setMotifs,pvalSimulThresh,randomFile,fm);
            //Compute analytical p-values
            mm.computeAnalyticalPVals(edd,setMotifs,pvalAnalThresh,approx);
        }

        //Print results
        if(outputFile!=null)
        {
            fm.writeResults(outputFile,setMotifs,directed,true,false);
            System.out.println("Results written to "+outputFile);
        }
        else
            fm.printResults(setMotifs,directed,true);


        //Create Precision-recall plot and write it to output file
        if(outputFilePlot!=null)
        {
            System.out.println("Making precision-recall plot... ");
            PlotManager pm=new PlotManager(netFile,maxMotifSize);
            double thresholdPVal=0.05;
            JFreeChart chart=pm.createPRPlot(setMotifs,thresholdPVal);
            fm.savePlotToPDFFile(chart,outputFilePlot);
            System.out.println("Done! Plot saved in "+outputFilePlot);
        }

        double fine=System.currentTimeMillis();
        double totalTime=(fine-inizio)/1000;

        System.out.println("\nTime elapsed: "+totalTime+" secs");
    }
}
